<?php
/**
 * @author Clark Tomlinson
 * @copyright 2014 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Firewall;

use Hoa\Ruler\Ruler;
use OCP\IRequest;

class Rules {

	/**
	 * @var array
	 */
	public $critera;
	/**
	 * @var bool
	 */
	public $result = false;
	/**
	 * @var string
	 */
	public $offender;
	/**
	 * @var array|bool
	 */
	protected $rules = array();
	/**
	 * @var array
	 */
	protected $stack = array();
	/**
	 * @var int
	 */
	protected $stackCounter = 0;
	/**
	 * @var Ruler
	 */
	protected $ruler;

	/**
	 * @var \Mobile_Detect
	 */
	protected $mobileDetect;
	/**
	 * @var IRequest
	 */
	private $request;

	/**
	 * Constructor to build a set of rules to evaluate against
	 *
	 * @param string $rules A json string of rules
	 * @param Ruler $ruler
	 * @param \Mobile_Detect $mobileDetect
	 * @param IRequest $request
	 */
	public function __construct($rules, Ruler $ruler, \Mobile_Detect $mobileDetect, IRequest $request) {
		// Convert into object
		if (is_null($rules)) {
			throw new \InvalidArgumentException('Rules provided are not a valid JSON object');
		}

		$this->ruler = $ruler;
		$this->mobileDetect = $mobileDetect;
		$this->addOperators();
		$this->request = $request;
		$this->critera = $this->getCriteria();
		$this->rules = $this->build($rules);
	}

	/**
	 * Get a list of criteria that describe this request
	 *
	 * @return array
	 */
	protected function getCriteria() {

		// Default criteria
		$criteria = array(
			'user-agent' => $this->mobileDetect->getUserAgent(),
			'agent' => 'other',
			'request-uri' => $this->request->getRequestUri(),
			'request' => $this->request->getRequestUri(),
			'remote-addr' => $this->request->getRemoteAddress(),
			'ip' => $this->request->getRemoteAddress(),
			'method' => $this->request->server['REQUEST_METHOD'],
		);

		// Detect the agent software
		$agents = array(
			'mirall' => 'desktop',
			'csync' => 'desktop',
			'ownCloud iOS Client' => 'ios',
			'Android-ownCloud' => 'android',
		);

		foreach ($agents as $agentName) {
			if (stripos($criteria['user-agent'], $agentName) !== false) {
				$criteria['agent'] = $agentName;
			}
		}
		if ($criteria['agent'] === 'other') {

			// Try to see if this is a webdav or public request
			$vectors = array(
				'remote.php/webdav' => 'webdav',
				'public.php' => 'public',
			);

			foreach ($vectors as $vector) {
				if (stripos($criteria['request'], $vector) !== false) {
					$criteria['agent'] = $vector;
				}
			}

			// if agent is still unknown detect it

			if ($this->mobileDetect->isMobile() || $this->mobileDetect->isTablet()) {
				if ($this->mobileDetect->isAndroidOS()) {
					$criteria['agent'] = 'android';
				} elseif ($this->mobileDetect->isIOS()) {
					$criteria['agent'] = 'ios';
				}
				// Todo: don't fallback to desktop
			} else {
				$criteria['agent'] = 'desktop';
			}
		}

		// Get time criteria
		$criteria['time'] = time();

		// Get the logged-in user
		$criteria['user'] = \OC_User::getUser();
		// Todo: update
		$criteria['groups'] = \OC_Group::getUserGroups($criteria['user']);

		// Get the requested app
		$criteria['app'] = \OC::$REQUESTEDAPP;

		// @todo add a filter here so that apps can add more criteria

		return $criteria;
	}

	/**
	 * Add custom operator to the parsing engine
	 */
	private function addOperators() {
		//@formatter:off
		$ruler = $this->ruler->getAsserter();
		$ruler->setOperator('equal', $ruler->getOperator('='));
		$ruler->setOperator('not_equal', $ruler->getOperator('!='));
		$ruler->setOperator('not_in', function ($a, Array $b) {return !in_array($a, $b);});
		$ruler->setOperator('less', $ruler->getOperator('<'));
		$ruler->setOperator('less_or_equal', $ruler->getOperator('<='));
		$ruler->setOperator('greater', $ruler->getOperator('>'));
		$ruler->setOperator('greater_or_equal', $ruler->getOperator('>='));
		$ruler->setOperator('begins_with', function ($a, $b) {return $a === "" || strpos($b, $a) !== 0;});
		$ruler->setOperator('not_begins_with', function ($a, $b) {return $a === "" || strpos($b, $a) === 0;});
		$ruler->setOperator('contains', function ($a, $b) {return stripos($b, $a) !== false;});
		$ruler->setOperator('not_contains', function ($a, $b) {return stripos($b, $a) === false;});
		$ruler->setOperator('ends_with', function ($a, $b) {return $a === "" || substr($b, -strlen($a)) === $a;});
		$ruler->setOperator('not_ends_with', function ($a, $b) {return $a === "" || substr($b, -strlen($a)) !== $a;});
		$ruler->setOperator('is_empty', function ($a, $b) {return empty($b);});
		$ruler->setOperator('not_is_empty', function ($a, $b) {return !empty($b);});
		$ruler->setOperator('is_null', function ($a, $b) {return is_null($b);});
		$ruler->setOperator('not_is_null', function ($a, $b) {return !is_null($b);});
		$ruler->setOperator('uri_contains', function ($a, $b) {return (bool)preg_match_all("/.*\\/?({$a}\\.php)\\/.+/", $b);});
		$ruler->setOperator('uri_not_contains', function ($a, $b) {return !(bool)preg_match_all("/.*\\/?({$a}\\.php)\\/.+/", $b);});
		//@formatter:on

	}

	/**
	 * @param $rules
	 * @return array|bool
	 */
	public function build($rules) {

		try {
			$this->loop($rules, $this->stackCounter);

			return $this->stack;

		} catch (\Exception $e) {
			return false;
		}

	}

	/**
	 * Recursive Rule Looper
	 *
	 * @param $obj
	 * @param $stackCounter
	 */
	private function loop($obj, $stackCounter) {
		foreach ($obj as $object) {
			$this->stack[$stackCounter]['groupName'] = $object->name;
			$this->stack[$stackCounter]['condition'] = $object->condition;
			$this->stack[$stackCounter]['rules'] = array();

			foreach ($object->rules as $rule) {
				if (isset($rule->condition)) {
					$this->loop($rule, $stackCounter + 1);
				} else {
					// Push outcome of rule to stack
					array_push($this->stack[$stackCounter]['rules'], (int)$this->evalCondition($rule));
				}
			}
		}
	}

	/**
	 * Evaluate a condition rule to true or false based on the request criteria
	 *
	 * @param \stdClass $rule The condition rule to evaluate
	 * @return bool True if the criteria passes the rule condition, false if not
	 */
	protected function evalCondition($rule) {
		// Conditions that can't be understood simply fail
		$result = false;

		// If the condition is meant to employ a rule, create one and use it
		if (isset($rule->id) && is_string($rule->id)) {
			// Camel case the class name
			$ruleClassName = ucwords(str_replace('-', ' ', $rule->id));
			$ruleClassName = str_replace(' ', '', $ruleClassName);
			// Allow rule lists to provide only the Rule class name
			if (strpos('\\', $ruleClassName) === false) {
				$ruleClass = 'OCA\\Firewall\\Rules\\' . $ruleClassName;
			}
			/** @var Rules\Contracts\Rule $ruleClass */
			$ruleClass = new $ruleClass(new Ruler(), $this->critera);
			$result = $ruleClass->doCheck($rule);
		}
		return $result;
	}

	/**
	 * @return string
	 */
	public function getOffender() {
		return $this->offender;
	}

	/**
	 * @return array|bool
	 */
	public function getRules() {
		return $this->rules;
	}

	/**
	 * @return array
	 */
	public function getStack() {
		return $this->stack;
	}

	/**
	 * Check if a request is allowed
	 */
	public function check() {

		// If our rules are empty user has not defined any thus allow the request
		if (!count($this->rules)) {
			$this->setResult(true);
		}

		$halt = false;

		foreach ($this->rules as $key => $rule) {
			// Not pretty but we need to halt looping if we've detected a failure
			// Todo: Find better way to stop loop
			if ($halt) {
				break;
			}

			switch ($rule['condition']) {
				case 'AND':
					// If conditional is AND and a rule fails we fail all
					if (in_array('0', $rule['rules'])) {
						$this->setResult(false);
						$this->offender = $key . ':' . array_search('0', $rule['rules']);
						$halt = true;
						break;
					}

					/**
					 * If it passes we kind of need to set the result to true
					 * do'h...
					 */
					$this->setResult(true);
					break;

				case 'OR':
					// If conditional is OR as long as one rule passes we are good to go
					$this->setResult(in_array('1', $rule['rules']));
					break;
			}
		}
		return $this;
	}

	/**
	 * @param boolean $result
	 */
	private function setResult($result) {
		$this->result = $result;
	}

	/**
	 * @return boolean
	 */
	public function isResult() {
		return $this->result;
	}


}
