<?php
/**
 * ownCloud
 *
 * @author Lukas Reschke <lukas@owncloud.com>
 * @copyright (C) 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\user_shibboleth\Controller;

use OCA\user_shibboleth\USER_SHIBBOLETH;
use OCP\AppFramework\Controller;
use OCP\AppFramework\Http;
use OCP\AppFramework\Http\DataResponse;
use OCP\IL10N;
use OCP\IRequest;
use OCP\Security\ISecureRandom;

/**
 * Class PasswordController is used to handle WebDAV passwords that can be used
 * in combination with sync clients that don't implement the Shib authentication.
 *
 * @package OCA\user_shibboleth\Controller
 */
class PasswordController extends Controller {

	protected $userId;
	protected $request;
	protected $secureRandom;
	protected $l10n;

	/**
	 * @param string $AppName
	 * @param string $UserId
	 * @param IRequest $request
	 * @param ISecureRandom $secureRandom
	 * @param IL10N $l10n
	 */
	public function __construct($AppName,
								$UserId,
								IRequest $request,
								ISecureRandom $secureRandom,
								IL10N $l10n) {
		parent::__construct($AppName, $request);

		$this->userId = $UserId;
		$this->request = $request;
		$this->secureRandom = $secureRandom;
		$this->l10n = $l10n;
	}

	/**
	 * AJAX handler for generating a new alternate WebDAV password
	 *
	 * @NoAdminRequired
	 * @return DataResponse
	 */
	public function getNewPassword() {
		$backend = new \OCA\user_shibboleth\USER_SHIBBOLETH();
		$passwordSetup = $backend->getWebDAVPasswordConfiguration();
		$result = [];
		$desiredLength = $passwordSetup[0];
		$desiredGroupLength = $passwordSetup[1];
		$numOfGroups = (int)($desiredLength / $desiredGroupLength);

		while($numOfGroups > 0) {
			$result[] = $this->secureRandom
				->getMediumStrengthGenerator()
				->generate(
					$desiredGroupLength,
					ISecureRandom::CHAR_DIGITS.ISecureRandom::CHAR_LOWER.ISecureRandom::CHAR_UPPER
				);
			$numOfGroups--;
		}

		// FIXME: Make non-static to be able to unit-test this properly
		$fullPassword = implode($result);
		if (USER_SHIBBOLETH::setWebDAVPassword($this->userId, $fullPassword)) {
			return new DataResponse($result);
		} else {
			return new DataResponse(['message' => 'Could not save generated password'], Http::STATUS_UNPROCESSABLE_ENTITY);
		}
	}


	/**
	 * AJAX handler for revoking an alternate non-shibboleth password
	 *
	 * @NoAdminRequired
	 * @return DataResponse
	 */
	public function revokePassword() {
		// FIXME: Make non-static to be able to unit-test this properly
		if (USER_SHIBBOLETH::clearWebDAVPassword($this->userId)) {
			return new DataResponse();
		} else {
			return new DataResponse(['message' => 'Could not revoke the password'], Http::STATUS_UNPROCESSABLE_ENTITY);
		}
	}
}
