<?php
/**
 * ownCloud
 *
 * @author Thomas Müller <deepdiver@owncloud.com>
 * @author Iago López Galeiras
 * @copyright (C) 2014 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

use OCA\user_shibboleth\USER_SHIBBOLETH;

class UserShibbolethTest extends \PHPUnit_Framework_TestCase {

	/**
	 * @var USER_SHIBBOLETH
	 */
	private $backend;

	/**
	 * @var string
	 */
	private $testUser;

	/**
	 * @var string
	 */
	private $testUser2;

	/**
	 * @var string
	 */
	private $testUser3;

	/**
	 * @var string
	 */
	private $currentUser;

	/**
	 * @var integer
	 */
	private $savedPasswordLength;

	/**
	 * @var integer
	 */
	private $savedGroupLength;

	public function setup() {
		\OC::$server->getConfig()->setAppValue("user_shibboleth", "shibboleth_active", true);

		$this->currentUser = uniqid('shib-user-') . '@example.com';
		$_SERVER["PHP_AUTH_USER"] = $this->currentUser;
		$_SERVER["eppn"] = $this->currentUser;
		$_SERVER["Shib-Session-ID"] = uniqid('shib');

		$this->backend = new USER_SHIBBOLETH();
		$this->testUser = "test_shib_user";
		$this->testUser2 = "test_shib_user2";
		$this->testUser3 = "test_shib_user3";

		$query = \OC_DB::prepare('INSERT INTO `oc_shibboleth_user` VALUES(?, 0, ?, NULL)');
		$query->execute(array($this->testUser, $this->testUser));
		$query->execute(array($this->testUser2, 'displayName'));
		$query->execute(array($this->testUser3, null));

		$this->savedPasswordLength = \OC::$server->getConfig()->getSystemValue('non_shib_webdav_password_length');
		$this->savedGroupLength = \OC::$server->getConfig()->getSystemValue('non_shib_webdav_password_group_length');
	}

	public function tearDown() {
		$query = \OC_DB::prepare("DELETE FROM `oc_shibboleth_user` WHERE `eppn` = ?");
		$query->execute(array($this->testUser));
		$query->execute(array($this->testUser2));
		$query->execute(array($this->testUser3));

		if (!is_null($this->savedPasswordLength))
			\OC::$server->getConfig()->setSystemValue('non_shib_webdav_password_length', $this->savedPasswordLength);
		else
			\OC::$server->getConfig()->deleteSystemValue('non_shib_webdav_password_length');

		if (!is_null($this->savedGroupLength))
			\OC::$server->getConfig()->setSystemValue('non_shib_webdav_password_group_length', $this->savedGroupLength);
		else
			\OC::$server->getConfig()->deleteSystemValue('non_shib_webdav_password_group_length');
	}

	public function testIsSessionActiveIfNotActive() {
		\OC::$server->getConfig()->setAppValue("user_shibboleth", "shibboleth_active", false);
		$result = $this->backend->isSessionActive();

		$this->assertFalse($result);

		$result = $this->backend->userExists($this->currentUser);
		$this->assertFalse($result);
	}

	public function testIsSessionActiveNoUserSet() {
		unset($_SERVER["PHP_AUTH_USER"]);
		unset($_SERVER["eppn"]);
		$result = $this->backend->isSessionActive();

		$this->assertFalse($result);

		$result = $this->backend->userExists($this->currentUser);
		$this->assertFalse($result);
	}

	public function testGetCurrentUserId() {
		$result = $this->backend->getCurrentUserId();

		$this->assertEquals($this->currentUser, $result);
	}

	public function testIsSessionActive() {
		$result = $this->backend->isSessionActive();
		$this->assertTrue($result);

		# We call isSessionActive twice to increase code coverage, the first
		# call will create the user if it doesn't exist and the second one just
		# updates its display name and email
		$result = $this->backend->isSessionActive();
		$this->assertTrue($result);

		$result = $this->backend->userExists($this->currentUser);
		$this->assertTrue($result);
	}

	public function testWebDAVPassword() {
		$result = $this->backend->checkIfUserHasWebDAVPassword($this->testUser);
		$this->assertFalse((bool)$result);

		$result = $this->backend->setWebDAVPassword($this->testUser, '1234');
		$this->assertTrue((bool)$result);

		$result = $this->backend->checkIfUserHasWebDAVPassword($this->testUser);
		$this->assertTrue((bool)$result);

		$result = $this->backend->checkPassword($this->testUser, '1234');
		$this->assertTrue((bool)$result);

		$result = $this->backend->checkPassword($this->testUser, '12345');
		$this->assertFalse((bool)$result);

		$result = $this->backend->clearWebDAVPassword($this->testUser);
		$this->assertTrue((bool)$result);

		$result = $this->backend->checkIfUserHasWebDAVPassword($this->testUser);
		$this->assertFalse((bool)$result);
	}

	public function testNonExistentUserWebDAVPasword() {
		$result = $this->backend->userExists('bogusUser');
		$this->assertFalse((bool)$result);

		$result = $this->backend->setWebDAVPassword('bogusUser', '1234');
		$this->assertFalse((bool)$result);

		$result = $this->backend->checkIfUserHasWebDAVPassword('bogusUser');
		$this->assertFalse((bool)$result);

		$result = $this->backend->clearWebDAVPassword('bogusUser');
		$this->assertFalse((bool)$result);

		$result = $this->backend->checkPassword('bogusUser', '12345');
		$this->assertFalse((bool)$result);
	}

	public function testGetDisplayNames() {
		$result = $this->backend->getDisplayNames();
		$this->assertEquals($this->testUser, $result[$this->testUser]);
		$this->assertEquals('displayName', $result[$this->testUser2]);
		$this->assertEquals($this->testUser3, $result[$this->testUser3]);
	}

	public function testSearchDisplayNames() {
		$result = $this->backend->getDisplayNames('name');
		$this->assertArrayHasKey($this->testUser2, $result);
		$this->assertEquals('displayName', $result[$this->testUser2]);
	}

	public function testGetUser() {
		$result = $this->backend->getDisplayName($this->testUser2);
		$this->assertEquals('displayName', $result);
		$result = $this->backend->getDisplayName($this->testUser3);
		$this->assertEquals($this->testUser3, $result);
    }

	public function testGetUsers() {
		$result = $this->backend->getUsers();
		$this->assertTrue(in_array($this->testUser, $result));
		$this->assertTrue(in_array($this->testUser2, $result));
		$this->assertTrue(in_array($this->testUser3, $result));
		$this->assertFalse(in_array('test_shib_user4', $result));
		$userCount = sizeof($result);

		$result = $this->backend->countUsers();
		$this->assertEquals($userCount, $result);

	}

	public function testGetWebDAVPasswordConfiguration() {
		\OC::$server->getConfig()->setSystemValue('non_shib_webdav_password_length', 16);
		\OC::$server->getConfig()->setSystemValue('non_shib_webdav_password_group_length', 2);

		$result = $this->backend->getWebDAVPasswordConfiguration();

		$this->assertEquals(16, $result[0]);
		$this->assertEquals(2, $result[1]);

		\OC::$server->getConfig()->setSystemValue('non_shib_webdav_password_length', 0);
		\OC::$server->getConfig()->setSystemValue('non_shib_webdav_password_group_length', 0);

		$result = $this->backend->getWebDAVPasswordConfiguration();

		$this->assertEquals(12, $result[0]);
		$this->assertEquals(4, $result[1]);
	}
}
